#!/usr/bin/python

#
# Given a grid of letters, finds all words which can be
# spelled by moving no more than one space per letter
#

#
# Ways to implement:
#
# - A: O(num_words)
#   - Load dict
#   - Foreach word...
#     - exhaustive search for that word
#     - if found, add to results
#
# - B O(gridsize * 9**(length of longest word))
#   - Load dict into hash
#   - Foreach letter in grid
#     - limit search depth to length of longest word in dict
#     - Is it a word?
#     - recursively check each neighbor square
#
# - C
#   - Load dict into hash
#   - Load grid into hash-ish structure
#     - A list of letters
#     - Each contains a list of neighbors
#     - Probably equivalent to method A
#

#
# I think I'll use method A.  Simple, and apparently relatively efficient.
#

class letter:

  def __init__(self, l):
    self.l = l
    self.neighbors = []

class grid:

  def __init__(self):
    self.letters = []
    self.letters_flat = []
    self.g = []

  def load(self, filename):
    import string

    self.letters = []

    try:
      fp = open(filename, "rb")

      foo = fp.readline()
      parts = string.split(foo, " ")
      wid = int(parts[0])
      hgt = int(parts[1])

      self.wid = wid
      self.hgt = hgt

      for i in range(hgt):
        line = fp.readline()
        if len(line) != wid+1:
          print "Bad row on grid: '%s'" % line
        else:
          row = []
          for l in line[:-1]:
            let = letter(l)
            row.append(let)
            self.letters_flat.append(let)
          self.letters.append(row)

      self.generate_neighbors()

    except Exception, msg:
      print "Can't load grid: %s" % msg
      return

  def generate_neighbors(self):
    for y in range(self.hgt):
      for x in range(self.wid):
        for v in [-1, 0, 1]:
          for u in [-1, 0, 1]:
            i=x+u
            j=y+v
            #print "N(%s,%s) -> (%s,%s) ?" % (x, y, i, j)
            if i >= 0  \
               and i <  self.wid \
               and j >= 0 \
               and j <  self.hgt:
                 self.letters[y][x].neighbors.append(self.letters[j][i])
                 #print "Neighbor: %s -> %s" % (self.letters[y][x].l,
                 #                              self.letters[j][i].l)

  def has_word(self, word, let=None):
    #print "has_word(%s)" % (word)
    # recursion found word
    if not word:
      return 1

    # search entire grid, or just one square?
    if let:
      letters = let.neighbors
    else:
      letters = self.letters_flat

    # check all possible routes
    for l in letters:
      #print "Checking (%s == %s)?" % (word[0], l.l)
      if word[0] == l.l:
        found = self.has_word(word[1:], l)
        if found:
          return 1
    return 0

def load_dict(filename):
  d = []

  fp = open(filename, "rb")
  for l in fp.readlines():
    d.append(l[:-1])

  return d


def run(args):
  if len(args) < 2:
    print "Usage: bawg dictfile boggleboard"
  else:
    dictfile = args[0]
    bawgfile = args[1]

  # load dict
  dict = load_dict(dictfile)
  result = []

  g = grid()
  g.load(bawgfile)

  for word in dict:
    if g.has_word(word):
      result.append(word)
      print word

  #for word in result:
  #  print word

  print "Found %s words." % (len(result))


if __name__ == "__main__":
  import sys
  run(sys.argv[1:])
